# -*- coding: utf-8 -*
"""Convert ISA-Tab to CEDAR JSON"""
from __future__ import absolute_import
import json
import logging
import os
from os import listdir
from os.path import isdir, join
from uuid import uuid4

from jsonschema import Draft4Validator, RefResolver
from jsonschema.exceptions import ValidationError

from isatools.io import isatab_parser


__author__ = 'agbeltran'


log = logging.getLogger('isatools')
CEDAR_SCHEMA_PATH = join(
    os.path.dirname(os.path.realpath(__file__)), "../resources/schemas/cedar")


class ISATab2CEDAR(object):

    def __init__(self, primary_source):
        self.primary_source = primary_source

    def createCEDARjson_folder(self, work_dir, json_dir, inv_identifier):
        log.info("Convert ISA datasets in folder ".format(work_dir))
        path = os.path.abspath(work_dir)
        folders = [f for f in listdir(path) if isdir(join(path, f))]

        for folder in folders:
            self.createCEDARjson(CEDAR_SCHEMA_PATH, join(path, folder),
                                 json_dir, inv_identifier)

    def createCEDARjson(self, work_dir, json_dir, inv_identifier):
        log.info("Converting ISA to CEDAR model for ".format(work_dir))
        schema_file = "investigation_template.json"
        with open(join(CEDAR_SCHEMA_PATH, schema_file)) as json_fp:
            schema = json.load(json_fp)
        if schema is None:
            raise IOError("Could not load schema from {}".format(
                join(CEDAR_SCHEMA_PATH, schema_file)))
        resolver = RefResolver(
            'file://'+join(CEDAR_SCHEMA_PATH, schema_file), schema)
        validator = Draft4Validator(schema, resolver=resolver)

        isa_tab = isatab_parser.parse(work_dir)

        if isa_tab is None:
            log.info("No ISAtab dataset found")
        else:
            if isa_tab.metadata != {}:
                investigationObject = dict([
                    ("@id",
                     "https://repo.metadatacenter.org/UUID/"+str(uuid4())),
                    ("_templateId", "http://example.org"),
                    ("@type",
                     "https://repo.metadatacenter.org/model/Investigation"),
                    ("@context", dict(
                        [
                            ("description",
                             "https://metadatacenter.org/schemas/description"),
                            ("title",
                             "https://metadatacenter.org/schemas/title"),
                            ("study",
                             "https://metadatacenter.org/schemas/study"),
                            ("submissionDate",
                             "https://metadatacenter.org/schemas/"
                             "submissionDate"),
                            ("_value", "https://schema.org/value"),
                            ("publicReleaseDate",
                             "https://metadatacenter.org/schemas/"
                             "publicReleaseDate"),
                            ("identifier",
                             "https://metadatacenter.org/schemas/identifier")
                        ]
                    )),
                    ("title", dict(
                        [("_value",
                          isa_tab.metadata['Investigation Title'])])),
                    ("description", dict([("_value", isa_tab.metadata[
                        'Investigation Description'])])),
                    ("identifier", dict([("_value", isa_tab.metadata[
                        'Investigation Identifier'])])),
                    ("submissionDate", dict([("_value", isa_tab.metadata[
                        'Investigation Submission Date'])])),
                    ("publicReleaseDate", dict([("_value", isa_tab.metadata[
                        'Investigation Public Release Date'])])),
                    ("study", self.createStudiesList(isa_tab.studies))
                ])
            else:
                investigationObject = dict([
                    ("@id",
                     "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                    ("_templateId", "http://example.org"),
                    ("@type", "https://repo.metadatacenter.org/model/"
                              "Investigation"),
                    ("@context", dict(
                        [
                            ("description",
                             "https://metadatacenter.org/schemas/description"),
                            ("title",
                             "https://metadatacenter.org/schemas/title"),
                            ("study",
                             "https://metadatacenter.org/schemas/study"),
                            ("submissionDate",
                             "https://metadatacenter.org/schemas/"
                             "submissionDate"),
                            ("_value", "https://schema.org/value"),
                            ("publicReleaseDate",
                             "https://metadatacenter.org/schemas/"
                             "publicReleaseDate"),
                            ("identifier",
                             "https://metadatacenter.org/schemas/identifier")
                        ]
                    )),
                    ("title", dict([("_value", "")])),
                    ("description", dict([("_value", "")])),
                    ("identifier", dict([("_value", "")])),
                    ("submissionDate", dict([("_value", "")])),
                    ("publicReleaseDate", dict([("_value", "")])),
                    ("study", self.createStudiesList(isa_tab.studies)),
                ])

            cedar_json = investigationObject
            try:
                investigation_identifier = \
                    isa_tab.metadata['Investigation Identifier']
            except KeyError:
                investigation_identifier = ""

            try:
                study_identifier = \
                    isa_tab.studies[0].metadata['Study Identifier']
            except KeyError:
                study_identifier = ""

            try:
                validator.validate(cedar_json, schema)
            except ValidationError as e:
                error_file_name = os.path.join(json_dir, "error.log")
                with open(error_file_name, "w") as errorfile:
                    errorfile.write(e.message)
                    errorfile.write(e.cause)
                    errorfile.close()

            if inv_identifier:
                file_name = os.path.join(
                    json_dir, investigation_identifier+".json")
            else:
                file_name = os.path.join(
                    json_dir, study_identifier+".json")
            with open(file_name, "w") as outfile:
                json.dump(cedar_json, outfile, indent=4, sort_keys=True)
                outfile.close()

            log.info("... conversion finished.")

    def createStudiesList(self, studies):
        json_list = []
        for study in studies:
            source_dict = self.createSources(study.nodes)
            sample_dict = self.createSamples(study.nodes)
            # data_dict = self.createDataFiles(study.nodes)
            json_item = dict([
                ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                ("@type", "https://repo.metadatacenter.org/model/Study"),
                ("title", dict([("_value", study.metadata['Study Title'])])),
                ("description",
                 dict([("_value", study.metadata['Study Description'])])),
                ("identifier",
                 dict([("_value", study.metadata['Study Identifier'])])),
                ("submissionDate",
                 dict([("_value", study.metadata['Study Submission Date'])])),
                ("publicReleaseDate", dict(
                    [("_value",
                      study.metadata['Study Public Release Date'])])),
                ("studyDesignType", dict([("_value", "")])),
                ("publication",
                 self.createStudyPublicationsList(study.publications)),
                ("contact", self.createStudyContactsList(study.contacts)),
                ("studyFactor", self.createStudyFactorsList(study.factors)),
                ("studyAssay", self.createStudyAssaysList(study.assays)),
                ("studyGroupPopulation",
                 self.createStudyGroupList(source_dict)),
                ("studyProtocol",
                 self.createStudyProtocolList(study.protocols)),
                ("process", self.createProcessList(
                    study.process_nodes, source_dict, sample_dict))
            ])
            json_list.append(json_item)
        return json_list

    def createStudyGroupList(self, source_dict):
        json_list = []
        json_item = dict([
            ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
            ("@type", "http://purl.obolibrary.org/obo/STATO_0000193"),
            ("name", dict([("_value", "population name")])),
            ("type", dict([(
                "_value", "http://bioportal.bioontology.org/ontologies/EFO/"
                "3232")])),
            ("selectionRule",  dict([("_value", "selection rule")])),
            ("studySubject", list(source_dict.values()))
            ])
        json_list.append(json_item)
        return json_list

    def createProcessList(self, process_nodes, source_dict, sample_dict):
        json_list = []
        for process_node_name in process_nodes:
            try:
                measurement_type = \
                    process_nodes[process_node_name].study_assay.metadata[
                        "Study Assay Measurement Type"]
            except Exception:
                measurement_type = ""

            try:
                platform = \
                    process_nodes[process_node_name].study_assay.metadata[
                        "Study Assay Technology Platform"]
            except Exception:
                platform = ""

            try:
                technology = \
                    process_nodes[process_node_name].study_assay.metadata[
                        "Study Assay Technology Type"]
            except Exception:
                technology = ""

            process_node = process_nodes[process_node_name]

            json_item = dict([
                    ("@id",
                     "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                    ("@type",
                     "https://repo.metadatacenter.org/model/Process"),
                    ("type", dict([("_value", process_node_name)])),
                    ("studyProtocol",
                        self.createExecuteStudyProtocol(
                            process_node_name, process_node)),
                    ("studyAssay",
                     [{"@type": "http://purl.obolibrary.org/obo/BFO_0000055",
                       "@id": "https://repo.metadatacenter.org/UUID",
                       "measurementType": {"_value": measurement_type},
                       "platform": {"_value": platform},
                       "technology": {"_value": technology}}]),
                    ("input", self.createInputList(
                        process_node.inputs, source_dict, sample_dict)),
                    ("output",
                        self.createOutputList(
                            process_node.outputs, sample_dict)),
                    ("parameterValue",
                     self.createParameterValueList(process_node))
            ])
            json_list.append(json_item)
        return json_list

    def createInputList(self, inputs, source_dict, sample_dict):
        json_dict = dict([])
        sample_list = []
        studySubject_list = []
        for argument in inputs:
            try:
                json_item = source_dict[argument]
                studySubject_list.append(json_item)
            except KeyError:
                pass
            try:
                json_item = sample_dict[argument]
                sample_list.append(json_item)
            except KeyError:
                pass
        json_dict.update({"sample": sample_list})
        json_dict.update({"studySubject": studySubject_list})
        return json_dict

    def createOutputList(self, arguments, sample_dict):
        json_dict = dict([])
        sample_list = []
        for argument in arguments:
            try:
                json_item = sample_dict[argument]
                sample_list.append(json_item)
            except KeyError:
                pass
        json_dict.update({"sample": sample_list})
        return json_dict

    def createExecuteStudyProtocol(self, process_node_name, process_node):
        json_item = dict([
                    ("@context", dict(
                            [
                                ("name",
                                 "https://metadatacenter.org/schemas/name"),
                                ("uRI",
                                 "https://metadatacenter.org/schemas/uRI"),
                                ("_value", "https://schema.org/value"),
                                ("version",
                                 "https://metadatacenter.org/schemas/version"),
                                ("protocolParameter",
                                 "https://metadatacenter.org/schemas/"
                                 "protocolParameter"),
                                ("type",
                                 "https://metadatacenter.org/schemas/type"),
                                ("description",
                                 "https://metadatacenter.org/schemas/"
                                 "description")
                            ]
                    )),
                    ("name", dict([("_value", process_node_name)])),
                    ("@id",
                     "https://repo.metadatacenter.org/UUID" + str(uuid4())),
                    ("@type", "http://purl.obolibrary.org/obo/OBI_0000272"),
                    ("type",
                     dict([("_value",
                            "http://purl.obolibrary.org/obo/OBI_0000272")])),
                    ("description", dict([("_value", process_node_name)])),
                    ("version", dict([("_value", process_node_name)])),
                    ("uRI", dict([("_value", process_node_name)])), (
                        "protocolParameter",
                        self.createProtocolParameterFromNode(process_node))])

        return [json_item]

    def createProtocolParametersList(self, protocol):
        json_list = []
        parameters = protocol['Study Protocol Parameters Name']
        parametersURIs = protocol[
            'Study Protocol Parameters Name Term Accession Number']
        index = 0
        if len(parameters) > 0:
            for parameter in parameters.split(';'):
                json_item = dict([
                    ("name", dict([("_value", parameter)])),
                    ("description",
                        (dict([("_value", parametersURIs[index] if (
                            len(parametersURIs) == len(
                                parameters)) else "")]))),
                ])
                index = index+1
                json_list.append(json_item)
        return json_list

    def createProtocolParameterFromNode(self, process_node):
        json_list = []
        json_item = dict([
            ("description", dict([("_value", process_node.protocol)])),
            ("name", dict([("_value", process_node.protocol)]))
        ])
        json_list.append(json_item)
        return json_list

    def createParameterValueList(self, process_node):
        json_list = []
        for header in process_node.metadata:
            value_header = header.replace("]", "").split("[")[-1]
            value_attributes = process_node.metadata[header][0]
            value = value_attributes[0]
            if header.startswith("Parameter Value"):
                json_item = dict([
                    ("@id",
                     "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                    ("@type", "https://repo.metadatacenter.org/model/"
                              "ProtocolParameter"),
                    ("value", dict([("_value", value)])),
                    ("protocolParameter",
                     self.createProtocolParameterFromNode(process_node)),
                    ("type", dict([("_value", value_header)])),
                    ("unit", dict([("_value", value_header)]))
                ])
                json_list.append(json_item)
        return json_list

    def createDataFiles(self, nodes):
        json_dict = dict([])
        for node_index in nodes:
            if nodes[node_index].ntype.endswith("Data File"):
                json_item = dict([
                    ("@id",
                     "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                    ("@type",
                     "https://repo.metadatacenter.org/model/DataFile"),
                    ("name", dict([("_value", node_index)])),
                    ("type", dict([(
                        "_value",
                        "http://purl.obolibrary.org/obo/OBI_0000747")])),
                    ("description", dict([("_value", "")]))
                ])
                json_dict.update({node_index: json_item})
        return json_dict

    def createSamples(self, nodes):
        json_dict = dict([])
        for node_index in nodes:
            if nodes[node_index].ntype == "Sample Name":
                json_item = dict([
                    ("@id",
                     "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                    ("@type", "https://repo.metadatacenter.org/model/Sample"),
                    ("name", dict([("_value", node_index)])),
                    ("type", dict([(
                        "_value",
                        "http://purl.obolibrary.org/obo/OBI_0000747")])),
                    ("description", dict([("_value", "")])),
                    ("source", dict([("_value", "")])),
                    ("factorValue",
                     self.createFactorValueList(nodes[node_index])),
                    ("studyTime", self.createStudyTimeCollection()),
                    ("characteristic",
                     self.createCharacteristicList(nodes[node_index])),
                ])
                json_dict.update({node_index: json_item})
        return json_dict

    def createSources(self, nodes):
        json_dict = dict([])
        for node_index in nodes:
            if nodes[node_index].ntype == "Source Name":
                json_item = dict([
                    ("name", dict([("_value", node_index)])),
                    ("type", dict([(
                        "_value",
                        "http://purl.obolibrary.org/obo/OBI_0000925")])),
                    ("characteristic",
                     self.createCharacteristicList(nodes[node_index])),
                ])
                json_dict.update({node_index: json_item})
        return json_dict

    def createStudyTimeCollection(self):
        json_item = dict([
                    ("@id",
                     "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                    ("@type", "http://purl.obolibrary.org/obo/OBI_0001619"),
                    ("durationValue", dict([("_value", "")])),
                    ("isBeforeEvent", dict([("_value", "")])),
                    ("studyEvent", dict([("_value", "")])),
                    ("unit", dict([("_value", "")]))
                  ])
        return json_item

    def createCharacteristicList(self, node):
        json_list = []
        for header in node.metadata:
            if header.startswith("Characteristics"):
                characteristic = header.replace("]", "").split("[")[-1]
                json_item = dict([
                    ("@id",
                     "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                    ("@type",
                     "https://repo.metadatacenter.org/model/Characteristic"),
                    ("name", dict([("_value", characteristic)])),
                    ("description", dict([("_value", "")])),
                    ("characteristicValue", self.createCharacteristicValueList(
                        node.metadata[header][0]))
                    ])
                json_list.append(json_item)
        return json_list

    def createCharacteristicValueList(self, value_attributes):

        value = value_attributes[0]
        try:
            typeValue = value_attributes.Term_Accession_Number
        except AttributeError:
            typeValue = ""

        try:
            unitValue = value_attributes.Unit
        except AttributeError:
            unitValue = ""

        if unitValue:
            characteristicValue = dict([
                ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                ("@type",
                 "https://repo.metadatacenter.org/model/CharacteristicValue"),
                ("type", dict([("_value", "")])),
                ("unit", dict([("_value", unitValue), ("@type", typeValue)])),
                ("value", dict([("_value", value)]))
                ])
        else:
            characteristicValue = dict([
                ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                ("@type",
                 "https://repo.metadatacenter.org/model/CharacteristicValue"),
                ("type", dict([("_value", typeValue)])),
                ("unit", dict([("_value", unitValue)])),
                ("value", dict([("_value", value)]))
                ])

        return characteristicValue

    def createFactorValueList(self, node):
        factor_list = []
        for header in node.metadata:
            if header.startswith("Factor Value"):
                value_header = header.replace("]", "").split("[")[-1]
                value_attributes = node.metadata[header][0]
                value = value_attributes[0]
                try:
                    unit = value_attributes.Unit
                except AttributeError:
                    unit = ""
                factorValue = dict([
                    ("@id",
                     "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                    ("@type", "https://repo.metadatacenter.org/model/"
                              "CharacteristicValue"),
                    ("type", dict([("_value", value_header)])),
                    ("unit", dict([("_value", unit), ("@type", "")])),
                    ("value", dict([("_value", value)])),
                    ("studyFactor", [self.createStudyFactor("", "")])
                    ])
                factor_list.append(factorValue)
        return factor_list

    def createInvestigationContactsList(self, contacts):
        json_list = []
        for contact in contacts:
            json_item = dict([
                ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                ("@type", "https://repo.metadatacenter.org/model/Contact"),
                ("lastName", dict([("_value", contact[
                    'Investigation Person Last Name'])])),
                ("firstName", dict([("_value", contact[
                    'Investigation Person First Name'])])),
                ("middleInitial", dict([("_value", contact[
                    'Investigation Person Mid Initials'])])),
                ("email", dict([("_value", contact[
                    'Investigation Person Email'])])),
                ("phone", dict([("_value", contact[
                    'Investigation Person Phone'])])),
                ("fax", dict([("_value", contact[
                    'Investigation Person Fax'])])),
                ("address", dict([("_value", contact[
                    'Investigation Person Address'])])),
                ("role", dict([("_value", contact[
                    'Investigation Person Roles Term Accession Number'])])),
                ("organization", self.createAffiliationsList(
                    contact['Investigation Person Affiliation']))
                ])
            json_list.append(json_item)
        return json_list

    def createStudyContactsList(self, contacts):
        json_list = []
        for contact in contacts:
            json_item = dict([
                ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                ("@type", "https://repo.metadatacenter.org/model/Contact"),
                ("lastName", dict([("_value", contact[
                    'Study Person Last Name'])])),
                ("firstName", dict([("_value", contact[
                    'Study Person First Name'])])),
                ("middleInitial", dict([("_value", contact[
                    'Study Person Mid Initials'])])),
                ("email", dict([("_value", contact['Study Person Email'])])),
                ("phone", dict([("_value", contact['Study Person Phone'])])),
                ("fax", dict([("_value", contact['Study Person Fax'])])),
                ("address", dict([("_value", contact[
                    'Study Person Address'])])),
                ("role", dict([("_value", contact[
                    'Study Person Roles Term Accession Number'])])),
                ("organization", self.createAffiliationsList(
                    contact['Study Person Affiliation']))
                ])
            json_list.append(json_item)
        return json_list

    def createInvestigationPublicationsList(self, publications):
        json_list = []
        for publication in publications:
            json_item = dict([
                ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                ("@type", "https://repo.metadatacenter.org/model/Publication"),
                ("title", dict([("_value", publication[
                    'Investigation Publication Title'])])),
                ("pubMedID", dict([("_value", publication[
                    'Investigation PubMed ID'])])),
                ("dOI", dict([("_value", publication[
                    'Investigation Publication DOI'])])),
                ("authorList", self.createAuthorList(
                    publication['Investigation Publication Author List'])),
                ("status", dict([("_value", publication[
                    'Investigation Publication Status'])])),
                ])
            json_list.append(json_item)
        return json_list

    def createStudyPublicationsList(self, publications):
        json_list = []
        for publication in publications:
            json_item = dict([
                ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                ("@type", "https://repo.metadatacenter.org/model/Publication"),
                ("title", dict([("_value", publication[
                    'Study Publication Title'])])),
                ("pubMedID", dict([("_value", publication[
                    'Study PubMed ID'])])),
                ("dOI", dict([("_value", publication[
                    'Study Publication DOI'])])),
                ("authorList", self.createAuthorList(
                    publication['Study Publication Author List'])),
                ("status", dict([("_value", publication[
                    'Study Publication Status'])])),
                ])
            json_list.append(json_item)
        return json_list

    def createAffiliationsList(self, affiliations):
        json_list = []
        json_item = dict([
                ("@context", ""),
                ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                ("@type", "https://metadatacenter.org/model/Organization"),
                ("name", dict([("_value", affiliations)])),
                ("department", dict([("_value", "")]))
                ])
        json_list.append(json_item)
        return json_list

    def createStudyAssaysList(self, assays):
        json_list = []
        for assay in assays:
            json_item = dict([
                ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                ("@type", "http://purl.obolibrary.org/obo/BFO_0000055"),
                ("measurementType", dict([("_value", assay.metadata[
                    'Study Assay Measurement Type Term Accession Number'])])),
                ("platform", dict([("_value", assay.metadata[
                    'Study Assay Technology Platform'])])),
                ("technology", dict([("_value", assay.metadata[
                    'Study Assay Technology Type'])]))
                ])
            json_list.append(json_item)
        return json_list

    def createStudyProtocolList(self, protocols):
        json_list = []
        for protocol in protocols:
            json_item = dict([
                ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
                ("@type",
                 "https://repo.metadatacenter.org/model/StudyProtocol"),
                ("name", dict([("_value", protocol['Study Protocol Name'])])),
                ("description", dict([("_value", protocol[
                    'Study Protocol Description'])])),
                ("type", dict([("_value", protocol['Study Protocol Type'])])),
                ("version", dict([("_value", protocol[
                    'Study Protocol Version'])])),
                ("uRI", dict([("_value", protocol['Study Protocol URI'])])),
                ("protocolParameter",
                 self.createProtocolParametersList(protocol)),
                ])
            json_list.append(json_item)
        return json_list

    def createStudyFactor(self, factor_name, factor_desc):
        json_item = dict([
            ("@id", "https://repo.metadatacenter.org/UUID"+str(uuid4())),
            ("@context", ""),
            ("@type", "http://www.ebi.ac.uk/efo/EFO_0000001"),
            ("name", dict([("_value", factor_name)])),
            ("description", dict([("_value", factor_desc)]))
        ])
        return json_item

    def createStudyFactorsList(self, factors):
        json_list = []
        for factor in factors:
            json_item = self.createStudyFactor(
                factor['Study Factor Name'], factor['Study Factor Type'])
            json_list.append(json_item)
        return json_list

    def createAuthorList(self, authorListString):
        json_list = []
        elements = authorListString.split(',')
        for element in elements:
            json_item = dict([
                ("_value", element)
            ])
            json_list.append(json_item)
        return json_list



# -*- coding: utf-8 -*-
"""Parse ISA-Tab structured metadata describing experimental data.
Works with ISA-Tab (http://isatab.sourceforge.net), which provides a structured
format for describing experimental metdata.
The entry point for the module is the parse function, which takes an ISA-Tab
directory (or investigator file) to parse. It returns a ISATabRecord object
which contains details about the investigation. This is top level information
like associated publications and contacts.
This record contains a list of associated studies (ISATabStudyRecord objects).
Each study contains a metadata attribute, which has the key/value pairs
associated with the study in the investigation file. It also contains other
high level data like publications, contacts, and details about the experimental
design.
The nodes attribute of each record captures the information from the Study
file.
This is a dictionary, where the keys are sample names and the values are
NodeRecord objects. This collapses the study information on samples, and
contains the associated information of each sample as key/value pairs in the
metadata attribute.
Finally, each study contains a list of assays, as ISATabAssayRecord objects.
Similar to the study objects, these have a metadata attribute with key/value
information about the assay. They also have a dictionary of nodes with data
from
the Assay file; in assays the keys are raw data files.
This is a biased representation of the Study and Assay files which focuses on
collapsing the data across the samples and raw data.
"""
from __future__ import with_statement
import bisect
import collections
import csv
import glob
import os
import pprint
import re
from io import StringIO

from isatools import isatab


__author__ = 'brad chapman, agbeltran, djcomlab'

# REGEXES
_RX_COLLAPSE_ATTRIBUTE = re.compile(r"[\W]+")

# column labels
_LABELS_MATERIAL_NODES = ['Source Name', 'Sample Name', 'Extract Name',
                          'Labeled Extract Name']
_LABELS_DATA_NODES = ['Raw Data File', 'Derived Spectral Data File',
                      'Derived Array Data File', 'Array Data File',
                      'Protein Assignment File', 'Peptide Assignment File',
                      'Post Translational Modification Assignment File',
                      'Acquisition Parameter Data File',
                      'Free Induction Decay Data File',
                      'Derived Array Data Matrix File', 'Image File',
                      'Derived Data File', 'Metabolite Assignment File']
_LABELS_ASSAY_NODES = ['Assay Name', 'MS Assay Name',
                       'Hybridization Assay Name', 'Scan Name',
                       'Data Transformation Name', 'Normalization Name']


def find_lt(a, x):
    """Find rightmost value less than x"""
    i = bisect.bisect_left(a, x)
    if i:
        return a[i - 1]
    else:
        return -1


def find_gt(a, x):
    """Find leftmost value greater than x"""
    i = bisect.bisect_right(a, x)
    if i != len(a):
        return a[i]
    else:
        return -1


def find_in_between(a, x, y):
    result = []
    while True:
        try:
            element_gt = find_gt(a, x)
        except ValueError:
            return result

        if (element_gt > x and y == -1) or (element_gt > x and element_gt < y):
            result.append(element_gt)
            x = element_gt
        else:
            break

    while True:
        try:
            element_lt = find_lt(a, y)
        except ValueError:
            return result
        if element_lt not in result:
            if (element_lt < y and element_lt > x):
                result.append(element_lt)
                y = element_lt
            else:
                break
        else:
            break

    return result


def find_between(a, x, y):
    """Find value a between x and y"""
    result = []
    i = bisect.bisect_right(a, x)
    if i != len(a):
        while i < len(a):
            result.append(i)
            i += 1
            return result
    raise ValueError


def parse(isatab_ref):
    """Entry point to parse an ISA-Tab directory.
    isatab_ref can point to a directory of ISA-Tab data, in which case we
    search for the investigator file, or be a reference to the high level
    investigation file.
    """
    if os.path.isdir(isatab_ref):
        fnames = glob.glob(os.path.join(isatab_ref, "i_*.txt")) + \
            glob.glob(os.path.join(isatab_ref, "*.idf.txt"))
        assert len(fnames) == 1
        isatab_ref = fnames[0]
    assert os.path.exists(isatab_ref), \
        "Did not find investigation file: %s" % isatab_ref
    i_parser = InvestigationParser()
    with open(isatab_ref, encoding='utf-8') as in_handle:
        rec = i_parser.parse(in_handle)
    s_parser = StudyAssayParser(isatab_ref)
    rec = s_parser.parse(rec)
    return rec


class InvestigationParser:
    """Parse top level investigation files into ISATabRecord objects.
    """

    def __init__(self):
        self._sections = {
            "ONTOLOGY SOURCE REFERENCE": "ontology_refs",
            "INVESTIGATION": "metadata",
            "INVESTIGATION PUBLICATIONS": "publications",
            "INVESTIGATION CONTACTS": "contacts",
            "STUDY DESIGN DESCRIPTORS": "design_descriptors",
            "STUDY PUBLICATIONS": "publications",
            "STUDY FACTORS": "factors",
            "STUDY ASSAYS": "assays",
            "STUDY PROTOCOLS": "protocols",
            "STUDY CONTACTS": "contacts"}
        self._nolist = ["metadata"]

    def parse(self, in_handle):
        line_iter = self._line_iter(in_handle)
        # parse top level investigation details
        rec = ISATabRecord()
        rec, _ = self._parse_region(rec, line_iter)

        # parse study information
        while True:
            study = ISATabStudyRecord()
            study, had_info = self._parse_region(study, line_iter)
            if had_info:
                rec.studies.append(study)
            else:
                break
        # handle SDRF files for MAGE compliant ISATab
        if "SDRF File" in rec.metadata:
            study = ISATabStudyRecord()
            study.metadata["Study File Name"] = rec.metadata["SDRF File"]
            rec.studies.append(study)
        return rec

    def _parse_region(self, rec, line_iter):
        """Parse a section of an ISA-Tab, assigning information to a
        supplied record.
        """
        had_info = False
        keyvals, section = self._parse_keyvals(line_iter)

        if keyvals:
            rec.metadata = keyvals[0]
        while section and section[0] != "STUDY":
            had_info = True
            keyvals, next_section = self._parse_keyvals(line_iter)
            attr_name = self._sections[section[0]]
            if attr_name in self._nolist:
                try:
                    keyvals = keyvals[0]
                except IndexError:
                    keyvals = {}

            setattr(rec, attr_name, keyvals)
            section = next_section
        return rec, had_info

    @staticmethod
    def _line_iter(in_handle):
        """Read tab delimited file, handling ISA-Tab special case headers.
        """
        reader = csv.reader(in_handle, dialect="excel-tab")
        for line in reader:
            if len(line) > 0 and line[0]:
                # check for section headers; all uppercase and a single value
                if line[0].upper() == line[0] and "".join(line[1:]) == "":
                    line = [line[0]]
                yield line

    @staticmethod
    def _parse_keyvals(line_iter):
        """Generate dictionary from key/value pairs.
        """
        out = None
        line = None
        for line in line_iter:
            if len(line) == 1 and line[0].upper() == line[0]:
                break
            else:
                # setup output dictionaries, trimming off blank columns
                if out is None:
                    while not line[-1]:
                        line = line[:-1]
                    out = [{} for _ in line[1:]]
                # add blank values if the line is stripped
                while len(line) < len(out) + 1:
                    line.append("")
                for i in range(len(out)):
                    out[i][line[0]] = line[i + 1].strip()
                line = None
        return out, line


class StudyAssayParser:
    """Parse row oriented metadata associated with study and assay samples.
    This currently does not attempt to be complete, but rather to extract the
    most useful information (in my biased opinion) and represent it simply
    in the record objects.
    This is coded generally, so can be expanded to more cases. It is biased
    towards microarray and next-gen sequencing data.
    """

    def __init__(self, base_file):
        self._dir = os.path.dirname(base_file)
        self._col_quals = ("Performer", "Date", "Unit",
                           "Term Accession Number", "Term Source REF")
        self._col_types = {"attribute": ("Characteristics", "Factor Type",
                                         "Comment", "Label", "Material Type",
                                         "Factor Value"),
                           "node": (
                               "Sample Name", "Source Name", "Image File",
            "Raw Data File", "Derived Data File",
            "Acquisition Parameter Data File",
            "Extract Name", "Labeled Extract Name",
            "Array Data File",
            "Raw Spectral Data File",
            "Protein Assignment File",
            "Peptide Assignment File",
            "Post Translational Modification Assignment File",
                               "Derived Spectral Data File",
            "Derived Array Data File"),
            "node_assay": (
                               "Assay Name",
                               "Data Transformation Name",
            "Normalization Name"),
            "processing": "Protocol REF",
            "parameter": ("Parameter Value", "Array Design REF")
        }
        self._synonyms = {
            "Hybridization Assay Name": "Assay Name",
            "Scan Name": "Assay Name",
            "MS Assay Name": "Assay Name"
        }

    def parse(self, rec):
        """Retrieve row data from files associated with the ISATabRecord.
        """
        final_studies = []
        for study in rec.studies:
            source_data = self._parse_study(
                study.metadata["Study File Name"],
                self._col_types["node"])
            # ["Source Name", "Sample Name", "Comment[ENA_SAMPLE]"])
            if source_data:
                study.nodes = source_data
                final_assays = []
                for assay in study.assays:
                    cur_assay = ISATabAssayRecord(assay)
                    assay_data = self._parse_study(
                        assay["Study Assay File Name"],
                        self._col_types["node"])
                    cur_assay.nodes = assay_data
                    assay_process_nodes = self._get_process_nodes(
                        assay["Study Assay File Name"], cur_assay)

                    cur_assay.process_nodes = assay_process_nodes
                    final_assays.append(cur_assay)
                study.assays = final_assays

                # get process nodes
                study_process_nodes = self._get_process_nodes(
                    study.metadata["Study File Name"], study)
                study.process_nodes = study_process_nodes
                final_studies.append(study)
        rec.studies = final_studies
        return rec

    def _get_process_nodes(self, fname, study):
        """Building the process nodes"""
        if not os.path.exists(os.path.join(self._dir, fname)):
            return {}
        process_nodes = {}

        with self._preprocess(os.path.join(self._dir, fname)) as in_handle:
            reader = csv.reader(in_handle, dialect="excel-tab")
            headers = self._swap_synonyms(next(reader))
            hgroups = self._collapse_header(headers)
            htypes = self._characterize_header(headers, hgroups)
            processing_indices = [i for i, x in enumerate(htypes)
                                  if x == "processing"]
            all_parameters_indices = [i for i, x in enumerate(htypes)
                                      if x == "parameter"]
            node_indices = [i for i, x in enumerate(htypes)
                            if x == "node"]
            node_assay_indices = [i for i, x in enumerate(htypes)
                                  if x == "node_assay"]
            line_number = 0
            # max_number = 0
            process_counters = {}
            assay_name_map = {}
            input_process_map = {}
            output_process_map = {}

            for line in reader:
                previous_processing_node = None
                for processing_index in processing_indices:

                    processing_name = line[hgroups[processing_index][0]]
                    if not processing_name:
                        continue
                    next_processing_index = find_gt(
                        processing_indices, processing_index)
                    previous_processing_index = find_lt(
                        processing_indices, processing_index)

                    input_indices = find_in_between(
                        node_indices, previous_processing_index,
                        processing_index)
                    output_indices = find_in_between(
                        node_indices, processing_index,
                        next_processing_index)
                    parameters_indices = find_in_between(
                        all_parameters_indices, processing_index,
                        next_processing_index)
                    assay_name_indices = find_in_between(
                        node_assay_indices, processing_index,
                        next_processing_index)
                    qualifier_indices = hgroups[processing_index][1:]

                    input_headers = [
                        headers[hgroups[x][0]]
                        for i, x in enumerate(input_indices)]
                    output_headers = [
                        headers[hgroups[x][0]]
                        for i, x in enumerate(output_indices)]
                    processing_header = headers[
                        hgroups[processing_index][0]]

                    # qualifier_headers = [
                    #     headers[x] for i, x in enumerate(
                    #         qualifier_indices)]
                    qualifier_values = [
                        line[x]
                        for i, x in enumerate(qualifier_indices)]

                    # parameters_values = [
                    #     line[x]
                    #     for i, x in enumerate(parameters_indices)]

                    input_values = [line[
                        hgroups[x][0]]
                        for i, x in enumerate(input_indices)]
                    input_node_indices = [
                        self._build_node_index(
                            input_headers[i], input_values[i])
                        for i, x in enumerate(input_values)]

                    output_values = [
                        line[hgroups[x][0]]
                        for i, x in enumerate(output_indices)]
                    output_node_indices = [self._build_node_index(
                        output_headers[i], output_values[i])
                        for i, x in enumerate(output_values)]

                    qualifier_indices_string = '-'.join(qualifier_values)
                    input_node_indices_string = \
                        "-".join(input_node_indices)
                    output_node_indices_string = \
                        "-".join(output_node_indices)
                    # parameters_indices_string = '-'.join(parameters_values)

                    assay_name = ""
                    if assay_name_indices:
                        if len(assay_name_indices) == 1:
                            assay_name = line[
                                hgroups[assay_name_indices[0]][0]]

                    if assay_name:
                        unique_process_name = assay_name
                    else:
                        try:
                            unique_process_name = \
                                input_process_map[
                                    qualifier_indices_string
                                    + input_node_indices_string]
                            if not (unique_process_name.startswith(
                                    processing_name)):
                                raise KeyError
                        except KeyError:
                            try:
                                unique_process_name = \
                                    output_process_map[
                                        qualifier_indices_string
                                        + output_node_indices_string]
                                if not (unique_process_name.startswith(
                                        processing_name)):
                                    raise KeyError
                            except KeyError:
                                try:
                                    process_number = \
                                        process_counters[processing_name]
                                except KeyError:
                                    process_number = 0

                                process_number += 1
                                process_counters.update(
                                    {processing_name: process_number})
                                unique_process_name = \
                                    processing_name + str(process_number)

                    try:
                        process_node = process_nodes[unique_process_name]
                    except KeyError:
                        # create process node
                        process_node = ProcessNodeRecord(
                            unique_process_name,
                            processing_header, study,
                            processing_name)

                    if previous_processing_node:
                        previous_processing_node.next_process = \
                            process_node
                        process_node.previous_process = \
                            previous_processing_node

                    previous_processing_node = process_node
                    # previous_protocol = line[
                    # hgroups[next_processing_index]]

                    if assay_name:
                        process_node.assay_name = assay_name
                        assay_name_map.update({assay_name: process_node})

                    # Add qualifiers (performer and date)
                    for qualifier_index in qualifier_indices:
                        qualifier_header = headers[qualifier_index]
                        if qualifier_header == "Date":
                            process_node.date = line[qualifier_index]
                        elif qualifier_header == "Performer":
                            process_node.performer = line[qualifier_index]

                    if not (input_node_indices in process_node.inputs):
                        in_first = set(process_node.inputs)
                        in_second = set(input_node_indices)
                        in_second_but_not_in_first = in_second - in_first
                        process_node.inputs = \
                            process_node.inputs + list(
                                in_second_but_not_in_first)
                    if not (output_node_indices in process_node.outputs):
                        in_first = set(process_node.outputs)
                        in_second = set(output_node_indices)
                        in_second_but_not_in_first = in_second - in_first
                        process_node.outputs = \
                            process_node.outputs + list(
                                in_second_but_not_in_first)

                    input_process_map[qualifier_indices_string
                                      + input_node_indices_string] = \
                        unique_process_name
                    output_process_map[qualifier_indices_string
                                       + output_node_indices_string] = \
                        unique_process_name

                    # Add parameters
                    for parameter_index in parameters_indices:
                        parameter_header = headers[hgroups[
                            parameter_index][0]]
                        process_node.parameters.append(parameter_header)
                        # creating the metadata object
                        process_node.metadata = collections.defaultdict(set)
                        attrs = self._line_keyvals(line, headers, hgroups,
                                                   htypes,
                                                   process_node.metadata)
                        process_node.metadata = attrs

                    # max_number = max(len(process_node.inputs),
                    # len(process_node.outputs))
                    line_number += 1
                    process_nodes[unique_process_name] = process_node
                else:
                    line_number += 1
                # study.process_nodes = process_nodes
        return dict([(k, self._finalize_metadata(v))
                     for k, v in process_nodes.items()])

    def _preprocess(self, fname):
        df = isatab.read_tfile(fname)
        df = isatab.preprocess(DF=df)
        out_handle = StringIO()
        df.to_csv(out_handle, header=df.isatab_header, sep='\t', index=False)
        out_handle.seek(0)
        return out_handle

    def _parse_study(self, fname, node_types):
        """Parse study or assay row oriented file around the supplied base
        node."""

        if not os.path.exists(os.path.join(self._dir, fname)):
            return None
        nodes = {}
        with self._preprocess(os.path.join(self._dir, fname)) as in_handle:
            reader = csv.reader(in_handle, dialect="excel-tab")
            headers = self._swap_synonyms(next(reader))
            hgroups = self._collapse_header(headers)
            htypes = self._characterize_header(headers, hgroups)

            node_indices = [i for i, x in enumerate(htypes) if x == "node"]
            all_attribute_indices = [i for i, x in enumerate(
                htypes) if x == "attribute"]

            for node_index in node_indices:

                node_type = headers[hgroups[node_index][0]]
                if node_type not in node_types:
                    continue
                try:
                    header_index = hgroups[node_index][0]

                except ValueError:
                    header_index = None

                if header_index is None:
                    # print "Could not find standard header name: %s in %s" \
                    #                        % (node_type, header)
                    continue

                next_node_index = find_gt(node_indices, node_index)
                previous_node_index = find_lt(node_indices, node_index)
                attribute_indices = find_in_between(
                    all_attribute_indices, node_index, next_node_index)

                in_handle.seek(0, 0)
                for line in reader:
                    if (line[0].startswith("#")):
                        continue
                    name = self._swap_synonyms([line[header_index]])[0]
                    # skip the header line and empty lines
                    if (not name or name in headers):
                        continue
                    # to deal with same name used for different node types
                    # (e.g. Source Name and Sample Name using the same string)
                    node_index_name = self._build_node_index(node_type, name)

                    try:
                        node = nodes[node_index_name]

                        for attribute_index in attribute_indices:
                            attribute_header = headers[
                                hgroups[attribute_index][0]]
                            if attribute_header.startswith(
                                    "Factor Value") \
                                    and node_type != "Sample Name":
                                continue
                            if attribute_header not in node.attributes:
                                node.attributes.append(attribute_header)

                    except KeyError:
                        node = NodeRecord(name, node_type, node_index_name)
                        node.metadata = collections.defaultdict(set)
                        nodes[node_index_name] = node
                        attrs = self._line_keyvals(
                            line, headers, hgroups, htypes, node.metadata)
                        nodes[node_index_name].metadata = attrs

                        for attribute_index in attribute_indices:
                            attribute_header = headers[
                                hgroups[attribute_index][0]]
                            if attribute_header.startswith(
                                    "Factor Value") \
                                    and node_type != "Sample Name":
                                continue
                            if attribute_header not in node.attributes:
                                node.attributes.append(attribute_header)

                    if not (previous_node_index == -1):
                        node.derivesFrom.append(line[previous_node_index])

        return dict([(k, self._finalize_metadata(v))
                     for k, v in nodes.items()])

    @staticmethod
    def _finalize_metadata(node):
        """Convert node metadata back into a standard dictionary and list.
        """
        final = {}
        for key, val in iter(node.metadata.items()):
            # val = list(val)
            # if isinstance(val[0], tuple):
            #    val = [dict(v) for v in val]
            final[key] = list(val)
        node.metadata = final
        return node

    def _line_keyvals(self, line, header, hgroups, htypes, out):
        out = self._line_by_type(line, header, hgroups, htypes, out, "node")
        out = self._line_by_type(
            line, header, hgroups, htypes, out, "attribute",
            self._collapse_attributes)
        out = self._line_by_type(
            line, header, hgroups, htypes, out, "processing",
            self._collapse_attributes)
        out = self._line_by_type(
            line, header, hgroups, htypes, out, "parameter",
            self._collapse_attributes)
        return out

    @staticmethod
    def _line_by_type(line, header, hgroups, htypes, out, want_type,
                      collapse_quals_fn=None):
        """Parse out key value pairs for line information based on a group of
        values."""
        for index, htype in ((i, t) for i, t in enumerate(
                htypes) if t == want_type):
            col = hgroups[index][0]
            key = header[col]
            if collapse_quals_fn:
                val = collapse_quals_fn(line, header, hgroups[index])
            else:
                val = line[col]
            out[key].add(val)
        return out

    def _collapse_attributes(self, line, header, indexes):
        """Combine attributes in multiple columns into single named tuple.
        """
        names = []
        vals = []
        for i in indexes:
            if header[i]:
                names.append(_RX_COLLAPSE_ATTRIBUTE.sub(
                    "_", self._clean_header(header[i])))
                vals.append(line[i])
        Attrs = collections.namedtuple('Attrs', names)
        return Attrs(*vals)

    @staticmethod
    def _clean_header(header):
        """Remove ISA-Tab specific information from Header[real name] headers.
        """
        if header.find("[") >= 0:
            header = header.replace("]", "").split("[")[-1]
        # ISATab can start with numbers but this is not supported in
        # the python datastructure, so prefix with isa_ to make legal
        try:
            int(header[0])
            header = "isa_" + header
        except ValueError:
            pass
        return header

    def _characterize_header(self, header, hgroups):
        """Characterize header groups into different data types.
        """
        out = []
        for h in [header[g[0]] for g in hgroups]:
            this_ctype = None
            for ctype, names in self._col_types.items():
                if (h in names) or (h.startswith(names) and h.endswith("]")):
                    this_ctype = ctype
                    break
            out.append(this_ctype)
        return out

    def _collapse_header(self, header):
        """Combine header columns into related groups.
        """
        out = []
        for i, h in enumerate(header):
            if h.startswith(self._col_quals):
                out[-1].append(i)
            else:
                out.append([i])
        return out

    def _swap_synonyms(self, header):
        return [self._synonyms.get(h, h) for h in header]

    # to ensure uniqueness of node indexes
    @staticmethod
    def _build_node_index(type, name):
        if type == "Source Name":
            return "source-" + name
        elif type == "Sample Name":
            return "sample-" + name
        elif type == "Extract Name":
            return "extract-" + name
        elif type == "Labeled Extract Name":
            return "labeledextract-" + name
        elif type == "Raw Data File":
            return "rawdatafile-" + name
        elif type == "Derived Data File":
            return "deriveddatafile-" + name
        elif type == "Acquisition Parameter Data File":
            return "acquisitionparameterfile-" + name
        elif type == "Image File":
            return "imagefile-" + name
        elif type == "Array Data File":
            return "arraydatafile-" + name
        elif type == "Array Data Matrix File":
            return "arraydatamatrixfile-" + name
        elif type == "Derived Array Data Matrix File":
            return "derivedarraydatamatrixfile-" + name
        elif type == "Raw Spectral Data File":
            return "rawspectraldatafile-" + name
        elif type == "Derived Array Data Matrix File":
            return "derivedarraydatamatrixfile-" + name
        elif type == "Protein Assignment File":
            return "proteinassignmentfile-" + name
        elif type == "Peptide Assignment File":
            return "peptideassignmentfile-" + name
        elif type == "Post Translational Modification Assignment File":
            return "posttranslationalmodificationassignmentfile-" + name
        elif type == "Free Induction Decay Data File":
            return "freeinductiondecaydatafile-" + name
        elif type == "Derived Array Data File":
            return "derivedarraydatafile-" + name
        elif type == "Derived Spectral Data File":
            return "derivedspectraldatafile-" + name
        else:
            print("ERROR - Type not being considered! ", type)
            return name


_record_str = \
    """* ISATab Record
 metadata: {md}
 studies:
{studies}
"""

_study_str = \
    """  * Study
   metadata: {md}
   design_descriptors: {design_descriptors}
   publications : {publications}
   factors: {factors}
   protocols: {protocols}
   nodes:
    {nodes}
   process_nodes:
    {process_nodes}
   assays:
{assays}
"""

_assay_str = \
    """    * Assay
     metadata: {md}
     nodes:
        {nodes}
     process_nodes:
       {process_nodes}
"""

_node_str = \
    """       * Node -> {name} {type} {index}
         attributes: {attributes}
         derivesFrom: {derivesFrom}
         metadata: {md}"""

_process_node_str = \
    """       * Process Node ->  {name} {type}
         assay_name: {assay_name}
         protocol: {protocol}
         performer: {performer}
         date: {date}
         inputs: {inputs}
         outputs: {outputs}
         previous_process = {previous_process}
         next_process = {next_process}
         parameters: {parameters}
         metadata: {md}"""


class ISATabRecord:
    """Represent ISA-Tab metadata in structured format.
    High level key/value data.
      - metadata -- dictionary
      - ontology_refs -- list of dictionaries
      - contacts -- list of dictionaries
      - publications -- list of dictionaries
    Sub-elements:
      - studies: List of ISATabStudyRecord objects.
    """

    def __init__(self):
        self.metadata = {}
        self.ontology_refs = []
        self.publications = []
        self.contacts = []
        self.studies = []

    def __str__(self):
        return _record_str.format(
            md=pprint.pformat(self.metadata).replace("\n", "\n" + " " * 3),
            ont=self.ontology_refs,
            pub=self.publications,
            contact=self.contacts,
            studies="\n".join(str(x) for x in self.studies))


class ISATabStudyRecord:
    """Represent a study within an ISA-Tab record.
    """

    def __init__(self):
        self.metadata = {}
        self.design_descriptors = []
        self.publications = []
        self.factors = []
        self.assays = []
        self.protocols = []
        self.contacts = []
        self.nodes = {}
        self.process_nodes = {}

    def __str__(self):
        return _study_str.format(
            md=pprint.pformat(self.metadata).replace("\n", "\n" + " " * 5),
            design_descriptors=pprint.pformat(
                self.design_descriptors).replace(
                "\n", "\n" + " " * 5),
            publications="\n".join(str(x) for x in self.publications),
            factors="\n".join(str(x) for x in self.factors),
            assays="\n".join(str(x) for x in self.assays),
            protocols="\n".join(str(x) for x in self.protocols),
            nodes="\n".join(str(x) for x in self.nodes.values()),
            process_nodes="\n".join(str(x)
                                    for x in self.process_nodes.values())
        )


class ISATabAssayRecord:
    """Represent an assay within an ISA-Tab record.
    """

    def __init__(self, metadata=None):
        if metadata is None:
            metadata = {}
        self.metadata = metadata
        self.nodes = {}
        self.process_nodes = {}

    def __str__(self):
        return _assay_str.format(
            md=pprint.pformat(self.metadata).replace("\n", "\n" + " " * 7),
            nodes="\n".join(str(x) for x in self.nodes.values()),
            process_nodes="\n".join(str(x)
                                    for x in self.process_nodes.values())
        )


class NodeRecord:
    """Represent a data or material node within an ISA-Tab Study/Assay file.
    """

    def __init__(self, name="", ntype="", nindex=""):
        self.ntype = ntype
        self.name = name
        self.index = nindex
        self.metadata = {}
        self.attributes = []
        self.derivesFrom = []

    def __str__(self):
        return _node_str.format(
            md=pprint.pformat(self.metadata).replace("\n", "\n" + " " * 9),
            index=self.index,
            name=self.name,
            type=self.ntype,
            attributes=pprint.pformat(
                self.attributes).replace(
                "\n", "\n" + " " * 9),
            derivesFrom=pprint.pformat(
                self.derivesFrom).replace("\n", "\n" + " " * 9))


class ProcessNodeRecord:
    """Represent a process node within an ISA-Tab Study/Assay file
    (corresponds to Protocol REF)."""

    def __init__(self, name="", ntype="", study_assay="", protocol=""):
        self.ntype = ntype
        self.study_assay = study_assay
        self.name = name
        self.protocol = protocol
        self.previous_process = None
        self.next_process = None
        self.inputs = []
        self.outputs = []
        self.metadata = {}
        self.parameters = []
        self.assay_name = ""
        # used when there is an associated 'Assay Name' for a 'Protocol REF'
        self.performer = ""
        self.date = ""

    def __str__(self):
        return _process_node_str.format(
            md=pprint.pformat(self.metadata).replace("\n", "\n" + " " * 9),
            inputs=pprint.pformat(self.inputs).replace("\n", "\n" + " " * 9),
            outputs=pprint.pformat(self.outputs).replace("\n", "\n" + " " * 9),
            name=self.name,
            assay_name=self.assay_name,
            type=self.ntype,
            protocol=self.protocol,
            performer=self.performer,
            date=self.date,
            previous_process=self.previous_process.name
            if self.previous_process else "",
            next_process=self.next_process.name if self.next_process else "",
            parameters=pprint.pformat(
                self.parameters).replace(
                "\n", "\n" + " " * 9)
        )


def strip_comments(in_fp):
    out_fp = StringIO()
    if not isinstance(in_fp, StringIO):
        out_fp.name = in_fp.name
    for line in in_fp.readlines():
        if line.lstrip().startswith('#'):
            pass
        else:
            out_fp.write(line)
    out_fp.seek(0)
    return out_fp



# -*- coding: utf-8 -*
"""
Convert ISA-Tab to W4M tables

original from
https://github.com/workflow4metabolomics/mtbls-dwnld/blob/develop/isatab2w4m.py
"""

# !/usr/bin/env python3
# vi: fdm=marker

import argparse
import collections
import csv
import glob
import logging
import os.path
import re
import sys
from string import Template

import numpy

from isatools import isatab as ISATAB
from isatools.utils import utf8_text_file_open

__author__ = 'pkrog (Pierrick Roger)'


log = logging.getLogger('isatools')

# Check Python version
if sys.hexversion < 0x03040000:
    sys.exit("Python 3.4 or newer is required to run this program.")


class FilenameTemplate(Template):
    delimiter = '%'


# Error message {{{1
################################################################

def error(msg):
    err = 'ERROR: {}'.format(msg)

    if __name__ == '__main__':
        print(err, file=sys.stderr)
        sys.exit(1)
    else:
        raise(IOError(err))


# Information message {{{1
################################################################

def info(msg):
    inf = 'INFO: {}'.format(msg)

    if __name__ == '__main__':
        print(inf)
    else:
        log.info(inf)


# Read arguments {{{1
################################################################

def read_args():
    s1 = 'You can use it as a template, where %%s will be replaced by the ' \
         'study name and %%a by the assay filename.'

    # Default values
    dft_output_dir = '.'
    dft_sample_file = '%s-%a-sample-metadata.tsv'
    dft_variable_file = '%s-%a-variable-metadata.tsv'
    dft_matrix_file = '%s-%a-sample-variable-matrix.tsv'

    parser = argparse.ArgumentParser(
        description='Script for extracting assays from ISATab data and '
                    'outputing in W4M format.')
    parser.add_argument('-a', help='Extract all assays.', dest='all_assays',
                        required=False, default=False, type=bool)
    parser.add_argument('-i',
                        help='Input directory containing the ISA-Tab files.',
                        dest='input_dir', required=True)
    parser.add_argument('-f',
                        help='Filename of the assay to extract. If unset, the '
                             'first assay of the chosen study will be used.',
                        dest='assay_filename', required=False)
    parser.add_argument('-n',
                        help='Filename of the study to extract. If unset, the '
                             'first study found will be used.',
                        dest='study_filename', required=False)
    parser.add_argument('-d',
                        help='Set output directory. Default is "' +
                             dft_output_dir + '".', dest="output_dir",
                        required=False, default=dft_output_dir)
    parser.add_argument('-s',
                        help='Output file for sample metadata. ' + s1 +
                             ' Default is "' + dft_sample_file.replace(
                                '%', '%%') + '".', dest="sample_output",
                        required=False, default=dft_sample_file)
    parser.add_argument('-v',
                        help='Output file for variable metadata. ' + s1 +
                             ' Default is "' + dft_variable_file.replace(
                                '%', '%%') + '".', dest='variable_output',
                        required=False, default=dft_variable_file)
    parser.add_argument('-m',
                        help='Output file for sample x variable matrix. ' +
                             s1 + ' Default is "' + dft_matrix_file.replace(
                                '%', '%%') + '".', dest='matrix_output',
                        required=False, default=dft_matrix_file)
    parser.add_argument('-S',
                        help='Filter out NA values in the specified sample '
                             'metadata columns. The value is a comma separated'
                             ' list of column names.',
                        dest='samp_na_filtering', required=False)
    parser.add_argument('-V',
                        help='Filter out NA values in the specified variable '
                             'metadata columns. The value is a comma separated'
                             ' list of column names.',
                        dest='var_na_filtering', required=False)
    args = parser.parse_args()
    args = vars(args)

    # Split comma separated list
    for opt in ['samp_na_filtering', 'var_na_filtering']:
        if opt in args and args[opt] is not None:
            args[opt] = args[opt].split(',')

    return args


# Select study {{{1
################################################################

def select_study(investigation_file, study_filename=None):
    investigation = load_investigation(investigation_file)
    study = None

    # More than one study and no study specified
    if len(investigation.studies) > 1 and study_filename is None:
        error(
            'The investigation file "{}" contains more than one study. You '
            'need to select one of them.'.format(investigation_file))

    # Search for specified study
    if study_filename is not None:

        # Loop on all studies
        for s in investigation.studies:
            if s.filename == study_filename:
                study = s
                break

        # Specified study not found
        if study is None:
            error(
                'Study "{0}" not found in investigation file "{1}".'
                .format(study_filename, investigation_file))

    # Take first one
    if study is None and len(investigation.studies) > 0:
        study = investigation.studies[0]

    return study


# Select assays {{{1
################################################################

def select_assays(study, assay_filename=None, all_assays=False):
    assays = []

    # Search for specified assay
    if assay_filename is not None and not all_assays:

        # Loop on all assays
        for a in study.assays:
            if a.filename == assay_filename:
                assays.append(a)
                break

        # Specified assay not found
        if len(assays) == 0:
            error('Assay "' + assay_filename + '" not found.')

    # Take all assays
    elif all_assays:
        assays = study.assays

    # Take first one
    else:
        assays = study.assays[0:1]

    return assays


# Get data file {{{1
################################################################

def get_data_file(assay):
    data_filename = None

    # Look for data files in assay
    for df in assay.data_files:
        m = re.match(r'^m_.*\.(tsv|txt)$', df.filename)
        if m is not None:
            if data_filename is not None:
                error('Found two data files ("{0}" and "{1}") in assay "{2}".'
                      .format(data_filename, df.filename, assay.filename))
            info('Found data file "' + df.filename + '".')
            data_filename = df.filename

    # No data file
    if data_filename is None:
        error('Found no data file in assay "{}".'.format(assay.filename))

    return data_filename


# Load data frame {{{1
################################################################

def load_df(path):
    df = ISATAB.read_tfile(path)
    df.replace(to_replace='', value=numpy.nan, inplace=True)
    return df


# Get assay data frame {{{1
################################################################

def get_assay_df(input_dir, assay):
    return load_df(os.path.join(input_dir, assay.filename))


# Get measures data frame {{{1
################################################################

def get_measures_df(input_dir, assay):
    data_filename = get_data_file(assay)
    return load_df(os.path.join(input_dir, data_filename))


# Get study data frame {{{1
################################################################

def get_study_df(input_dir, study):
    return load_df(os.path.join(input_dir, study.filename))


# Make names {{{1
################################################################

def make_names(u, uniq=False):
    v = u[:]
    j = 0
    for i in range(len(v)):

        # Create missing names
        if v[i] == '':
            v[i] = 'X' + ('' if j == 0 else ('.' + str(j)))
            j += 1

        # Remove unvanted characters
        else:
            v[i] = re.sub(r'[^A-Za-z0-9_.]', '.', v[i])

    # Make sure all elements are unique
    if uniq:
        # List all indices of items
        item_indices = collections.defaultdict(list)
        for i, x in enumerate(v):
            item_indices[x].append(i)

        # Look for duplicates
        for x in item_indices.keys():

            # Is this item duplicated?
            if len(item_indices[x]) > 1:

                # Rename all duplicates
                j = 1
                for i in item_indices[x][1:]:
                    while True:
                        new_name = x + "." + str(j)
                        if new_name not in item_indices.keys():
                            break
                        j += 1
                    v[i] = new_name
                    j += 1

    return v


# Make variable names {{{1
################################################################

def make_variable_names(assay_df):

    var_names = [''] * assay_df.shape[0]

    # Make variable names from data values
    for col in ['mass_to_charge', 'retention_time', 'chemical_shift']:
        if col in assay_df.keys():
            for i, v in enumerate(assay_df[col].values):
                if type(v) == str or not numpy.isnan(v):
                    x = var_names[i]
                    if x == '':
                        x = str(v)
                    else:
                        x = '_'.join([x, str(v)])
                    var_names[i] = x

    # Normalize names
    var_names = ['X' + s for s in var_names]
    var_names = make_names(var_names, uniq=True)

    return var_names


# Get investigation file {{{1
################################################################

def get_investigation_file(input_dir):
    # Search for file
    investigation_files = glob.glob(os.path.join(input_dir, 'i_*.txt'))

    # No file
    if len(investigation_files) == 0:
        error('No investigation file found.')

    # More than one file
    if len(investigation_files) > 1:
        error('Found more than one investigation file.')

    # File found
    investigation_file = investigation_files[0]
    info('Found investigation file "{}".'.format(investigation_file))

    return investigation_file


# Load investigation {{{1
################################################################

def load_investigation(investigation_file):
    f = utf8_text_file_open(investigation_file)
    investigation = ISATAB.load(f)
    return investigation


# Get sample names {{{1
################################################################

def get_sample_names(assay_df, measures_df):

    sample_names = None
    measures_cols = measures_df.axes[1]

    # Loop on all assay data frame columns
    for col in assay_df.axes[1]:
        n = assay_df.get(col).tolist()

        # Do you find all values of this column inside the column names of the
        # measure data frame?
        if len(n) == len(set(n)) and all([x in measures_cols for x in n]):
            sample_names = n
            break

    return sample_names


# Make sample metadata {{{1
################################################################

def make_sample_metadata(study_df, assay_df, sample_names, normalize=True):
    # Normalize column names
    study_df.set_axis(
        inplace=True, axis=1, labels=make_names(study_df.axes[1].tolist()))
    assay_df.set_axis(
        inplace=True, axis=1, labels=make_names(assay_df.axes[1].tolist()))

    # Merge data frames
    sample_metadata = assay_df.merge(study_df, on='Sample.Name', sort=False)

    # Normalize
    if normalize:
        norm_sample_names = make_names(sample_names, uniq=True)
        sample_metadata.insert(0, 'sample.name', norm_sample_names)
        sample_metadata.set_axis(inplace=True, axis=1, labels=make_names(
            sample_metadata.axes[1].tolist(), uniq=True))

    return sample_metadata


# Make variable metadata
################################################################

def make_variable_metadata(measures_df, sample_names, variable_names,
                           normalize=True):
    # Get variable columns from measures data frame
    all_cols = measures_df.axes[1].tolist()
    variable_cols = [x for x in all_cols if x not in sample_names]
    variable_metadata = measures_df.get(variable_cols)

    # Add variable names as columns
    variable_metadata.insert(0, 'variable.name', variable_names)

    # Normalize
    if normalize:
        variable_metadata.set_axis(inplace=True, axis=1, labels=make_names(
            variable_metadata.axes[1].tolist(), uniq=True))

    return variable_metadata


# Make matrix {{{1
################################################################

def make_matrix(measures_df, sample_names, variable_names, normalize=True):
    # Take all sample columns from measures data frame
    sample_variable_matrix = measures_df.get(sample_names)

    # Check that we got all columns
    if sample_variable_matrix is None or len(
            sample_variable_matrix.axes[1]) != len(sample_names):
        raise Exception(
            'Some or all sample names were not found among the column names of'
            ' the data array.')

    # Add variable names as columns
    sample_variable_matrix.insert(0, 'variable.name', variable_names)

    # Normalize sample names
    if normalize:
        norm_sample_names = make_names(sample_names, uniq=True)
        norm_sample_names.insert(0, 'variable.name')
        sample_variable_matrix.set_axis(
            inplace=True, axis=1, labels=norm_sample_names)

    return sample_variable_matrix


# Convert to W4M {{{1
################################################################

def convert2w4m(input_dir, study_filename=None, assay_filename=None,
                all_assays=False):
    # Select study
    investigation_file = get_investigation_file(input_dir)
    study = select_study(investigation_file, study_filename)
    if study is None:
        info('No studies found in investigation file.')
        return
    info('Processing study "{}".'.format(study.filename))

    # Select assays
    assays = select_assays(study=study, assay_filename=assay_filename,
                           all_assays=all_assays)

    # Loop on all assays
    w4m_assays = []
    for assay in assays:
        info('Processing assay "{}".'.format(assay.filename))
        study_df = get_study_df(input_dir, study)
        assay_df = get_assay_df(input_dir, assay)
        measures_df = get_measures_df(input_dir, assay)
        variable_names = make_variable_names(measures_df)
        sample_names = get_sample_names(assay_df=assay_df,
                                        measures_df=measures_df)
        sample_metadata = make_sample_metadata(study_df=study_df,
                                               assay_df=assay_df,
                                               sample_names=sample_names,
                                               normalize=True)
        variable_metadata = make_variable_metadata(
            measures_df=measures_df, sample_names=sample_names,
            variable_names=variable_names, normalize=True)
        sample_variable_matrix = make_matrix(measures_df=measures_df,
                                             sample_names=sample_names,
                                             variable_names=variable_names,
                                             normalize=True)
        w4m_assays.append(dict(samp=sample_metadata, var=variable_metadata,
                               mat=sample_variable_matrix,
                               filename=assay.filename,
                               study=study.identifier))

    return w4m_assays


# Write data frame {{{1
################################################################

def write_data_frame(df, output_dir, template_filename, study, assay):

    # NA values are removed by `read_tfile()` and replaced by ''.
    # Put them back here.
    df_with_na = df.replace(to_replace='', value=numpy.nan)

    # Set filename
    filename = FilenameTemplate(template_filename).substitute(s=study, a=assay)
    if output_dir is not None:
        filename = os.path.join(output_dir, filename)

    # Write data frame
    df_with_na.to_csv(
        path_or_buf=filename, sep='\t', na_rep='NA', index=False,
        quoting=csv.QUOTE_NONNUMERIC)


# Write assays into files {{{1
################################################################

def write_assays(assays, output_dir, samp_file, var_file, mat_file):
    # Create output directory if necessary
    if output_dir is not None and not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Make dict for file names
    filenames = dict(samp=samp_file, var=var_file, mat=mat_file)

    # Loop on all assays
    for assay in assays:
        for df in ['samp', 'var', 'mat']:
            write_data_frame(df=assay[df], output_dir=output_dir,
                             template_filename=filenames[df],
                             study=assay['study'], assay=assay['filename'])


# Filter NA values {{{1
################################################################

def filter_na_values(assays, samp_na_filtering=None, var_na_filtering=None):
    # Loop on all assays
    for assay in assays:

        if samp_na_filtering is not None:
            cols = make_names(samp_na_filtering)
            samp = assay['samp'].dropna(axis=0, how='all', subset=cols)
            removed_sample_names = numpy.setdiff1d(
                assay['samp']['sample.name'], samp['sample.name'])
            assay['samp'] = samp
            assay['mat'].drop(
                labels=removed_sample_names.tolist(), axis=1, inplace=True)

        if var_na_filtering is not None:
            cols = make_names(var_na_filtering)
            var_names = assay['var']['variable.name'].tolist()
            assay['var'].dropna(axis=0, how='all', subset=cols, inplace=True)
            kept_var_names = assay['var']['variable.name'].tolist()
            removed_variable_names_index = []
            for i, v in enumerate(var_names):
                if v not in kept_var_names:
                    removed_variable_names_index.append(i)
            assay['mat'].drop(labels=[
                assay['mat'].axes[0][i] for i in removed_variable_names_index],
                axis=0, inplace=True)


# Convert {{{1
################################################################

def convert(input_dir, output_dir, sample_output, variable_output,
            matrix_output, study_filename=None, assay_filename=None,
            all_assays=None, samp_na_filtering=None, var_na_filtering=None):
    # Convert assays to W4M format
    assays = convert2w4m(input_dir=input_dir,
                         study_filename=study_filename,
                         assay_filename=assay_filename,
                         all_assays=all_assays)

    # Filter NA values
    filter_na_values(assays, samp_na_filtering, var_na_filtering)

    # Write assays into files
    write_assays(assays, output_dir=output_dir, samp_file=sample_output,
                 var_file=variable_output, mat_file=matrix_output)


# Main {{{1
################################################################

def main():
    # Parse command line arguments
    args_dict = read_args()

    # Convert
    convert(input_dir=args_dict['input_dir'],
            output_dir=args_dict['output_dir'],
            variable_output=args_dict['variable_output'],
            matrix_output=args_dict['matrix_output'],
            sample_output=args_dict['sample_output'],
            study_filename=args_dict['study_filename'],
            assay_filename=args_dict['assay_filename'],
            all_assays=args_dict['all_assays'],
            samp_na_filtering=args_dict['samp_na_filtering'],
            var_na_filtering=args_dict['var_na_filtering']
            )


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*
"""Convert ISA-Tab to MAGE-TAB"""
import logging

from isatools import isatab, magetab


log = logging.getLogger('isatools')


def convert(source_inv_fp, output_path):
    """ Converter for ISA-Tab to MAGE-TAB.
    :param source_inv_fp: File descriptor of input investigation file
    :param output_dir: Path to directory to write output MAGE-TAB files to
    """
    log.info("loading isatab %s", source_inv_fp.name)
    ISA = isatab.load(source_inv_fp)
    log.info("dumping magetab %s", output_path)
    magetab.dump(ISA, output_path)



# -*- coding: utf-8 -*-
"""Functions for retrieving metadata from ArrayExpress.

This module connects to the European Bioinformatics Institute's
ArrayExpress database. If you have problems with it, check that
it's working at http://www.ebi.ac.uk/arrayexpress/
"""
from __future__ import absolute_import
import csv
import ftplib
import logging
import os
import shutil
import tempfile
import traceback
from pathlib import Path
from isatools.convert import magetab2isatab, magetab2json


EBI_FTP_SERVER = 'ftp.ebi.ac.uk'
AX_EXPERIMENT_BASE_DIR = '/pub/databases/arrayexpress/data/experiment'


log = logging.getLogger('isatools')


def get(arrayexpress_id, target_dir=None):
    """
    This function downloads MAGE-TAB content from the ArrayExpress FTP site.

    :param ax_experiment_id: Experiment identifier for ArrayExpress study to
    get, as a str (e.g. E-GEOD-59671)
    :param target_dir: Path to write MAGE-TAB files to. If None, writes to
    temporary directory (generated on the fly)
    :return: Path where the files were written to

    Example usage:
        from isatools.io import ax as AX
        AX.get('E-GEOD-59671', '/tmp/ax')
    """

    idbits = arrayexpress_id.split('-')
    exp_type = idbits[1]

    log.info("Setting up ftp with {}".format(EBI_FTP_SERVER))
    ftp = ftplib.FTP(EBI_FTP_SERVER)
    log.info("Logging in as anonymous user...")
    response = ftp.login()

    if '230' in response:  # 230 means Login successful
        log.info("Log in successful!")
        try:
            logging.info("Looking for experiment '{}'".format(arrayexpress_id))
            ftp.cwd('{base_dir}/{exp_type}/{arrayexpress_id}'.format(
                base_dir=AX_EXPERIMENT_BASE_DIR, exp_type=exp_type,
                arrayexpress_id=arrayexpress_id))
            if target_dir is None:
                target_dir = tempfile.mkdtemp()
                log.info("Using directory '{}'".format(target_dir))
            idf_filename = "{}.idf.txt".format(arrayexpress_id)
            with open(os.path.join(target_dir, idf_filename),
                      'wb') as out_file:
                log.info("Retrieving file '{}'".format(
                    EBI_FTP_SERVER + AX_EXPERIMENT_BASE_DIR + '/' + exp_type +
                    '/' + arrayexpress_id + '/' + idf_filename))
                ftp.retrbinary('RETR ' + idf_filename, out_file.write)
            try:
                with open(os.path.join(
                        target_dir, idf_filename),
                        encoding='utf-8') as unicode_idf_file:
                    reader = csv.reader(filter(
                        lambda r: r.startswith('SDRF File'), unicode_idf_file),
                        dialect='excel-tab')
                    for line in reader:
                        for sdrf_filename in line[1:]:
                            with open(os.path.join(
                                    target_dir, sdrf_filename),
                                    'wb') as out_file:
                                log.info("Retrieving file '{}'".format(
                                    EBI_FTP_SERVER +
                                    AX_EXPERIMENT_BASE_DIR +
                                    '/' + exp_type +
                                    '/' + arrayexpress_id +
                                    '/' + sdrf_filename))
                                ftp.retrbinary(
                                    'RETR ' + sdrf_filename, out_file.write)
            except UnicodeDecodeError:
                with open(os.path.join(
                        target_dir, idf_filename),
                        encoding='ISO8859-2') as latin2_idf_file:
                    reader = csv.reader(filter(
                        lambda r: r.startswith('SDRF File'),
                        latin2_idf_file), dialect='excel-tab')
                    for line in reader:
                        for sdrf_filename in line[1:]:
                            with open(os.path.join(
                                    target_dir, sdrf_filename),
                                    'wb') as out_file:
                                log.info("Retrieving file '{}'".format(
                                    EBI_FTP_SERVER + AX_EXPERIMENT_BASE_DIR +
                                    '/' + exp_type + '/' + arrayexpress_id +
                                    '/' + sdrf_filename))
                                ftp.retrbinary('RETR ' + sdrf_filename,
                                               out_file.write)
        except ftplib.error_perm as ftperr:
            log.fatal(
                "Could not retrieve ArrayExpress study '{study}': {error}"
                .format(study=arrayexpress_id, error=ftperr))
        finally:
            ftp.close()
            return target_dir
    else:
        ftp.close()
        raise ConnectionError(
            "There was a problem connecting to ArrayExpress: " + response)


def get_isatab(arrayexpress_id, target_dir=None):
    """
    This function downloads MAGE-TAB content as ISA-Tab from the
    ArrayExpress FTP site.

    :param ax_experiment_id: Experiment identifier for ArrayExpress study to
    get, as a str (e.g. E-GEOD-59671)
    :param target_dir: Path to write ISA-Tab files to. If None, writes to
    temporary directory (generated on the fly)
    :return: Path where the files were written to

    Example usage:
        from isatools.io import ax as AX
        AX.get_isatab('E-GEOD-59671', '/tmp/ax')
    """
    tmp_dir = tempfile.mkdtemp()
    try:
        tmp_dir = get(arrayexpress_id=arrayexpress_id, target_dir=tmp_dir)
        if target_dir is None:
            target_dir = tempfile.mkdtemp()
            log.info("Using directory '{}'".format(target_dir))
        fp = Path(os.path.join(tmp_dir, "{}.idf.txt".format(arrayexpress_id)))
        if fp.is_file():
            with fp.open('rb') as f:
                log.info("File {} content: {}".format(fp.absolute(), f.read()))
        else:
            log.critical("File {} does not exist!!".format(fp.absolute()))
        magetab2isatab.convert(os.path.join(tmp_dir, "{}.idf.txt".format(
            arrayexpress_id)), output_path=target_dir)
    except Exception as e:
        log.fatal("Something went wrong: {}".format(e))
        log.fatal(traceback.format_exc())
    finally:
        shutil.rmtree(tmp_dir)
        return target_dir


def getj(arrayexpress_id):
    """
    This function downloads MAGE-TAB content as ISA-JSON from the ArrayExpress
    FTP site.

    :param ax_experiment_id: Experiment identifier for ArrayExpress study to
    get, as a str (e.g. E-GEOD-59671)
    :return: ISA-JSON representation of the MAGE-TAB content

    Example usage:
        from isatools.io import ax as AX
        my_json = AX.getj('E-GEOD-59671')
    """
    tmp_dir = tempfile.mkdtemp()
    mage_json = None
    try:
        get(arrayexpress_id=arrayexpress_id, target_dir=tmp_dir)
        mage_json = magetab2json.convert(os.path.join(
            tmp_dir, "{}.idf.txt".format(arrayexpress_id)))
    except Exception as e:
        log.fatal("Something went wrong: {}".format(e))
    finally:
        shutil.rmtree(tmp_dir)
        return mage_json



# -*- coding: utf-8 -*
"""Convert SampleTab to ISA-Tab"""
import logging

from isatools import isatab, sampletab


logger = logging.getLogger('isatools')


def convert(source_sampletab_fp, target_dir):
    """ Converter for ISA-JSON to SampleTab.
    :param source_sampletab_fp: File descriptor of input SampleTab file
    :param target_dir: Path to write out ISA-Tab files to
    """
    ISA = sampletab.load(source_sampletab_fp)
    isatab.dump(ISA, target_dir)



# -*- coding: utf-8 -*
"""Convert ISA-JSON to ISA-Tab"""
import logging
import os
import shutil

from isatools import isajson, isatab


log = logging.getLogger('isatools')


def convert(json_fp, path, i_file_name='i_investigation.txt',
            config_dir=isajson.default_config_dir,
            validate_first=True):
    """
    Converter for ISA JSON to ISA Tab. Currently only converts
    investigation file contents
        :param json_fp: File pointer to ISA JSON input
        :param path: Directory to ISA tab output
        :param i_file_name: Investigation file name, default is
                            i_investigation.txt
        :param config_dir: Directory to config directory
        :param validate_first: Validate JSON before conversion, default is True

    Example usage:
        Read from a JSON and write to an investigation file, make sure to
        create/open relevant Python file objects.

        from isatools.convert import json2isatab
        json_file = open('BII-I-1.json', 'r')
        tab_file = open('i_investigation.txt', 'w')
        json2isatab.convert(json_file, path)

    """
    if validate_first:
        log.info("Validating input JSON before conversion")
        report = isajson.validate(fp=json_fp, config_dir=config_dir,
                                  log_level=logging.ERROR)
        if len(report['errors']) > 0:
            log.fatal("Could not proceed with conversion as there are some "
                      "fatal validation errors. Check log.")
            return
        json_fp.seek(0)  # reset file pointer after validation
    log.info("Loading ISA-JSON from %s", json_fp.name)
    isa_obj = isajson.load(fp=json_fp)
    log.info("Dumping ISA-Tab to %s", path)
    log.debug("Using configuration from %s", config_dir)
    isatab.dump(isa_obj=isa_obj, output_path=path, i_file_name=i_file_name)
    #  copy data files across from source directory where JSON is located
    log.info("Copying data files from source to target")
    for file in [f for f in os.listdir(os.path.dirname(json_fp.name))
                 if not (f.endswith('.txt') and (f.startswith('i_') or
                                                 f.startswith('s_') or
                                                 f.startswith('a_'))) and
                 not (f.endswith('.json'))]:
        filepath = os.path.join(os.path.dirname(json_fp.name), file)
        if os.path.isfile(filepath):
            log.debug("Copying %s to %s", filepath, path)
            shutil.copy(filepath, path)
